#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include "api.h"
#include "Alg.h"

#define CTESTS 10000
uint64_t loop = 1;

static void print_uint64(unsigned long long num)
{
	if (num >= 10)
		print_uint64(num / 10);
	printf("%u", (unsigned int)(num % 10));
}

int error_bit_num(unsigned char *k1, unsigned char *k2, int num)
{
	int i, sum = 0;
	unsigned char temp;
	for (i = 0; i < num; i++)
	{
		temp = k1[i] ^ k2[i];
		if (temp > 0)
		{
			sum += (temp & 0x1);
			sum += ((temp >> 1) & 0x1);
			sum += ((temp >> 2) & 0x1);
			sum += ((temp >> 3) & 0x1);
			sum += ((temp >> 4) & 0x1);
			sum += ((temp >> 5) & 0x1);
			sum += ((temp >> 6) & 0x1);
			sum += ((temp >> 7) & 0x1);
		}
	}

	return sum;
}
int print_error_bit(unsigned char *k1, unsigned char *k2, int num)
{
	int i, sum = 0;
	unsigned char temp;
	printf("\nerror bit:\n");
	for (i = 0; i < num; i++)
	{
		temp = k1[i] ^ k2[i];
		printf("%d%d%d%d%d%d%d%d", temp & 0x1, ((temp >> 1) & 0x1), ((temp >> 2) & 0x1), ((temp >> 3) & 0x1), ((temp >> 4) & 0x1), ((temp >> 5) & 0x1), ((temp >> 6) & 0x1), ((temp >> 7) & 0x1));
	}
	printf("\n");

	return sum;
}

//print bytes
int print_bytes(unsigned char *buf, int len)
{
	int i;
	for (i = 0; i < len; i++)
	{
		printf("%d ", buf[i]);
	}
	printf("\n");

	return 0;
}

//test correctness of pke
int test_pke_correctness()
{

	unsigned char pk[CRYPTO_PUBLICKEYBYTES];
	unsigned char sk[CRYPTO_SECRETKEYBYTES];
	unsigned char k1[CRYPTO_BYTES], k2[CRYPTO_BYTES], c[CRYPTO_CIPHERTEXTBYTES];
	int i, j;
	long long int  error_bit, sum = 0;
	long long int  error_num = 0, sum_bits;
	unsigned long long pk_byts=0, sk_byts=0, m_byts=CRYPTO_BYTES, c_byts=0;

	printf("correctness test of pke:\n");
	for (j = 0; j < loop; j++)
	{
		pke_keygen(pk, &pk_byts, sk, &sk_byts);
		rand_byts(CRYPTO_BYTES, k1);
		for (i = 0; i < CTESTS; i++)
		{

			pke_enc(pk, pk_byts, k1, m_byts, c, &c_byts);
			pke_dec(sk, sk_byts, c, c_byts, k2, &m_byts);
			//printf("i=%d\n",i);
			if (memcmp(k1, k2, m_byts) != 0)
			{
				error_num++;
				error_bit = error_bit_num(k1, k2, m_byts);
				sum += error_bit;
				if (error_bit > 0)
				{
					printf("error bit num:");
					print_uint64(error_bit);
					printf("\n");
					print_error_bit(k1, k2, m_byts);
				}
			}
		}
		printf("test %d error block:", j + 1);
		print_uint64(error_num);
		printf(" error bit:");
		print_uint64(sum);
		printf("\n");

	}
	sum_bits = CTESTS * loop*CRYPTO_BYTES * 8;
	printf("total error bit:");
	print_uint64(sum);
	printf("/");
	print_uint64(sum_bits);
	printf("\n\n");

	return error_num;
}

//test kem fo correctness
int test_kem_correctness()
{
	unsigned char pk[CRYPTO_PUBLICKEYBYTES];
	unsigned char sk[CRYPTO_SECRETKEYBYTES];
	unsigned char k1[CRYPTO_BYTES], k2[CRYPTO_BYTES], c[CRYPTO_CIPHERTEXTBYTES];
	int i, j;
	long long int  error_num = 0;
	unsigned long long pk_byts, sk_byts, ss_byts, c_byts;

	printf("correctness test of kem_fo:\n");
	for (j = 0; j < loop; j++)
	{
		kem_keygen(pk, &pk_byts, sk, &sk_byts);
		//rand_byts(CRYPTO_BYTES, k1);
		for (i = 0; i < CTESTS; i++)
		{
			kem_enc(pk, pk_byts, k1, &ss_byts, c, &c_byts);
			kem_dec(sk, sk_byts, c, c_byts, k2, &ss_byts);

			if (memcmp(k1, k2, CRYPTO_BYTES) != 0)
			{
				error_num++;
			}

		}
		printf("test %d error block:", j + 1);
		print_uint64(error_num);
		printf("\n");
	}
	printf("\n");

	return error_num;
}

//test ake correctness
int test_ake_correctness()
{
	unsigned char pka[CRYPTO_PUBLICKEYBYTES], pkb[CRYPTO_PUBLICKEYBYTES];
	unsigned char ska[CRYPTO_SECRETKEYBYTES], skb[CRYPTO_SECRETKEYBYTES], pa[CRYPTO_SECRETKEYBYTES+ CRYPTO_SECRETKEYBYTES + CRYPTO_BYTES + CRYPTO_CIPHERTEXTBYTES  + CRYPTO_PUBLICKEYBYTES], pb[CRYPTO_SECRETKEYBYTES];
	unsigned char k1[CRYPTO_BYTES], k2[CRYPTO_BYTES], ma[CRYPTO_PUBLICKEYBYTES + CRYPTO_CIPHERTEXTBYTES], mb[CRYPTO_CIPHERTEXTBYTES+CRYPTO_CIPHERTEXTBYTES];
	int i, j;
	long long int  error_num = 0;
	unsigned long long pka_byts, ska_byts, pkb_byts, skb_byts, pa_byts, pb_byts, ma_byts, mb_byts, k1_byts, k2_byts;

	printf("correctness test of ake:\n");
	for (j = 0; j < loop; j++)
	{
		for (i = 0; i < CTESTS; i++)
		{
			kex_init_a(pka, &pka_byts, ska, &ska_byts, pa, &pa_byts);
			kex_init_b(pkb, &pkb_byts, skb, &skb_byts, pb, &pb_byts);
			kex_send_a(pkb, pkb_byts, pa, &pa_byts, ma, &ma_byts);
			
			kex_send_b(pka, pka_byts, ma, ma_byts, pb, &pb_byts, mb, &mb_byts, k1, &k1_byts);
			kex_resp_a(mb, mb_byts, pa, &pa_byts, ma, &ma_byts, k2, &k2_byts);
			if (memcmp(k1, k2, CRYPTO_BYTES) != 0)
			{
				error_num++;
			}
		}
		printf("test %d error block:", j + 1);
		print_uint64(error_num);
		printf("\n");
	}
	printf("\n");

	return error_num;
}
