#include <stdio.h>
#include <string.h>
#include <time.h>
#include "test_speed.h"
#include "api.h"

#ifdef  __cplusplus
extern "C" {
#endif
#include "stdio.h"
	//#include "test_correctness.h"
#ifdef  __cplusplus
}
#endif

#define NTESTS 10000

	int init() {
		unsigned char pk[CRYPTO_PUBLICKEYBYTES];
		unsigned char sk[CRYPTO_SECRETKEYBYTES];
		int i;
		unsigned long long pk_byts, sk_byts;
		for (i = 0; i < NTESTS * 10; i++) {
			pke_keygen(pk, &pk_byts, sk, &sk_byts);
		}
		return 0;
	}
	//test pke
	int test_pke_speed()
	{
		clock_t start, finish;
		double total_time;
		unsigned char pk[CRYPTO_PUBLICKEYBYTES];
		unsigned char sk[CRYPTO_SECRETKEYBYTES];
		unsigned char k1[CRYPTO_BYTES], k2[CRYPTO_BYTES], c[CRYPTO_CIPHERTEXTBYTES];
		int i;
		unsigned long long mlen = CRYPTO_BYTES, clen = CRYPTO_CIPHERTEXTBYTES;
		unsigned long long pk_byts, sk_byts;



		start = clock();
		for (i = 0; i < NTESTS; i++)
		{
			pke_keygen(pk, &pk_byts, sk, &sk_byts);
		}
		finish = clock();
		total_time = (double)(finish - start) / CLOCKS_PER_SEC;
		printf("key generate time:%f us\n", (total_time * 1000000 / NTESTS));

		pke_keygen(pk, &pk_byts, sk, &sk_byts);
		rand_byts(CRYPTO_BYTES, k1);

		start = clock();
		for (i = 0; i < NTESTS; i++)
		{
			pke_enc(pk, pk_byts, k1, mlen, c, &clen);
		}
		finish = clock();
		total_time = (double)(finish - start) / CLOCKS_PER_SEC;
		printf("encryption time  :%f us\n", (total_time * 1000000 / NTESTS));

		pke_keygen(pk, &pk_byts, sk, &sk_byts);
		rand_byts(CRYPTO_BYTES, k1);
		pke_enc(pk, pk_byts, k1, mlen, c, &clen);

		start = clock();
		for (i = 0; i < NTESTS; i++)
		{
			pke_dec(sk, sk_byts, c, clen, k2, &mlen);
		}
		finish = clock();
		total_time = (double)(finish - start) / CLOCKS_PER_SEC;
		printf("decryption time  :%f us\n", (total_time * 1000000 / NTESTS));
		printf("\n");

		return 0;
	}
	
	//test kem fo
	int test_kem_speed()
	{
		clock_t start,finish;
		double total_time;
		unsigned char pk[CRYPTO_PUBLICKEYBYTES];
		unsigned char sk[CRYPTO_SECRETKEYBYTES];
		unsigned char k1[CRYPTO_BYTES],k2[CRYPTO_BYTES],c[CRYPTO_CIPHERTEXTBYTES];
		unsigned long long pk_byts, sk_byts, ss_byts, c_byts;
		int i;

		start=clock();
		for(i=0;i<NTESTS;i++)
		{
			kem_keygen(pk, &pk_byts, sk, &sk_byts);

		}
		finish=clock();
		total_time=(double)(finish-start)/CLOCKS_PER_SEC;
		printf("key generate time:%f us\n",(total_time*1000000/NTESTS));

		kem_keygen(pk, &pk_byts, sk, &sk_byts);

		start=clock();
		for(i=0;i<NTESTS;i++)
		{
			kem_enc(pk, pk_byts, k1, &ss_byts, c, &c_byts);
		}
		finish=clock();
		total_time=(double)(finish-start)/CLOCKS_PER_SEC;
		printf("kem_fo_enc time  :%f us\n",(total_time*1000000/NTESTS));

		kem_keygen(pk, &pk_byts, sk, &sk_byts);
		kem_enc(pk, pk_byts, k1, &ss_byts, c, &c_byts);

		start=clock();
		for(i=0;i<NTESTS;i++)
		{
			kem_dec(sk, sk_byts, c, c_byts, k2, &ss_byts);
		}
		finish=clock();
		total_time=(double)(finish-start)/CLOCKS_PER_SEC;
		printf("kem_fo_dec time  :%f us\n",(total_time*1000000/NTESTS));
		printf("\n");

		return 0;
	}
	
	// test ke
	int test_ke_speed()
	{
		clock_t start, finish;
		double total_time;
		unsigned char pka[CRYPTO_PUBLICKEYBYTES], pkb[CRYPTO_PUBLICKEYBYTES];
		unsigned char ska[CRYPTO_SECRETKEYBYTES], skb[CRYPTO_SECRETKEYBYTES], pa[CRYPTO_SECRETKEYBYTES], pb[CRYPTO_SECRETKEYBYTES];
		unsigned char k1[CRYPTO_BYTES], k2[CRYPTO_BYTES], ma[CRYPTO_CIPHERTEXTBYTES], mb[CRYPTO_CIPHERTEXTBYTES];
		unsigned long long pka_byts, ska_byts, pkb_byts, skb_byts, pa_byts, pb_byts, ma_byts, mb_byts, k1_byts, k2_byts;
		int i;

		//test alice send
		start = clock();
		for (i = 0; i < NTESTS; i++)
		{
			kex_init_a(pka, &pka_byts, ska, &ska_byts, pa, &pa_byts);
			kex_init_b(pkb, &pkb_byts, skb, &skb_byts, pb, &pb_byts);
			kex_send_a(pkb, pkb_byts, pa, &pa_byts, ma, &ma_byts);
			kex_send_b(pka, pka_byts, ma, ma_byts, pb, &pb_byts, mb, &mb_byts, k1, &k1_byts);
			kex_resp_a(mb, mb_byts, pa, &pa_byts, ma, &ma_byts, k2, &k2_byts);
		}
		finish = clock();
		total_time = (double)(finish - start) / CLOCKS_PER_SEC;
		printf("ke total time   :%f us\n", (total_time * 1000000 / NTESTS));


		return 0;
	}
	